@@ -32,7 +32,7 @@
  * optionally configure it with soup_logger_set_request_filter(),
  * soup_logger_set_response_filter(), and soup_logger_set_printer(),
  * and then attach it to a session (or multiple sessions) with
- * soup_logger_attach().
+ * soup_session_add_feature().
  *
  * By default, the debugging output is sent to %stdout, and looks
  * something like:
@@ -66,6 +66,17 @@
  * across the lifetime of the #SoupLogger. In particular, this can be
  * used to identify when multiple messages are sent across the same
  * connection.
+ *
+ * The request half of the message is logged just before the first
+ * byte of the request gets written to the network (from the
+ * #SoupSession::request_started signal), and the response is logged
+ * just after the last byte of the response body is read from the
+ * network (from the #SoupMessage::got_body or
+ * #SoupMessage::got_informational signal). In particular, note that
+ * the #SoupMessage::got_headers signal, and anything triggered off it
+ * (such as #SoupSession::authenticate) will be emitted
+ * <emphasis>before</emphasis> the response headers are actually
+ * logged.
  **/
 
 static void soup_logger_session_feature_init (SoupSessionFeatureInterface *feature_interface, gpointer interface_data);
@@ -79,7 +90,7 @@ static void request_unqueued  (SoupSessionFeature *feature,
 
 G_DEFINE_TYPE_WITH_CODE (SoupLogger, soup_logger, G_TYPE_OBJECT,
 			 G_IMPLEMENT_INTERFACE (SOUP_TYPE_SESSION_FEATURE,
-						soup_logger_session_feature_init));
+						soup_logger_session_feature_init))
 
 typedef struct {
 	/* We use a mutex so that if requests are being run in
@@ -476,8 +487,12 @@ print_request (SoupLogger *logger, SoupMessage *msg,
 	if (log_level == SOUP_LOGGER_LOG_MINIMAL)
 		return;
 
-	soup_logger_print (logger, SOUP_LOGGER_LOG_HEADERS, '>',
-			   "Host: %s", uri->host);
+	if (!soup_message_headers_get_one (msg->request_headers, "Host")) {
+		soup_logger_print (logger, SOUP_LOGGER_LOG_HEADERS, '>',
+				   "Host: %s%c%u", uri->host,
+				   soup_uri_uses_default_port (uri) ? '\0' : ':',
+				   uri->port);
+	}
 	soup_message_headers_iter_init (&iter, msg->request_headers);
 	while (soup_message_headers_iter_next (&iter, &name, &value)) {
 		if (!g_ascii_strcasecmp (name, "Authorization") &&
